[Version = "1.1.0"]
section AutodeskDataConnector;

// Forge OAuth2 values
client_id = "XXR6bKa3fHueh5WMygDxgMmDH0KEMGPO";
redirect_uri = "https://bim360dc-p-ue1.api.cosv2-p-ue1.autodesk.com";
base_path = "https://developer.api.autodesk.com";
prefix = "/data-connector/v1/accounts/";
accounts_prefix = "/bim360/";
authorize_uri = base_path & "/authentication/v1/authorize";
scope = "data:read data:write data:create account:read";

// Data Source Kind description
AutodeskDataConnector = [
    TestConnection = (dataSource) => { "AutodeskDataConnector.Contents" },
    Authentication = [
        OAuth = [
            StartLogin = StartLogin,
            FinishLogin = FinishLogin
        ]
    ],
    Label = Extension.LoadString("DataSourceLabel")
];

[DataSource.Kind="AutodeskDataConnector", Publish="AutodeskDataConnector.Publish"]

shared AutodeskDataConnector.Contents = Value.ReplaceType(AutodeskDataConnectorImpl, AutodeskDataConnectorType);

getHeader = (region as text) => 
    let 
        headerRegion = if (region = "US") then "US" else "EMEA"
    in 
        [
            Origin = "PowerBI Desktop",
            #"x-ads-region" = headerRegion
        ];       


AutodeskDataConnectorType = type function (
    region as (type text meta [
        Documentation.FieldCaption = "Region",
        // Documentation.FieldDescription = "",
        Documentation.AllowedValues = { "US", "EU" }
    ])
) as table meta [
        Documentation.Name = "Select Your Region: ",
        Documentation.LongDescription = "xxx " & Extension.LoadString("Version")
     ];

AutodeskDataConnectorImpl = (
     region as text
    ) =>
    let
        header = getHeader(region),
        accounts =  getAccounts(header),
        accountsNavigation = generateAccountsNavigation(accounts, header)
    in
        accountsNavigation;

getAccounts = (header) =>
    let
        response = Json.Document(
            Web.Contents(base_path & accounts_prefix & "admin/v1/accounts",
                [
                    Query = [ limit="200", fields="name"],
                    Headers = header
                ]
            )
        ),
        accounts = response[results]
    in
        accounts;

/**
 * Generate a NavigationTable for the accounts
 */
 
generateAccountsNavigation = (accounts, header) =>
    let
        #"Converted to Table" = Table.FromList(accounts, Splitter.SplitByNothing(), null, null, ExtraValues.Error),
        #"Expanded Column1" = Table.ExpandRecordColumn(#"Converted to Table", "Column1", {"name", "id"}, {"Name", "Key"}),
        #"Added Data" = Table.AddColumn(
            #"Expanded Column1", "Data", each generateFileTables([Key], header)
        ),
        #"Added ItemKind" = Table.AddColumn(#"Added Data", "ItemKind", each "Folder", Text.Type),
        #"Added ItemName" = Table.AddColumn(#"Added ItemKind", "ItemName", each "Folder", Text.Type),
        #"Added IsLeaf" = Table.AddColumn(#"Added ItemName", "IsLeaf", each false, Logical.Type),
        asNav = Table.ToNavigationTable(#"Added IsLeaf", {"Name"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf")
    in
        asNav;

generateFileTables = (id, header) =>
    let
        values = getViews(id, header),
        #"Added Data" = Table.AddColumn(
            values[data], "Data", each getCsv(values[account], values[job], [Name], header)
        ),
        #"Added ItemKind" = Table.AddColumn(#"Added Data", "ItemKind", each "Table", Text.Type),
        #"Added ItemName" = Table.AddColumn(#"Added ItemKind", "ItemName", each "Table", Text.Type),
        #"Added IsLeaf" = Table.AddColumn(#"Added ItemName", "IsLeaf", each true, Text.Type),
        asNav = Table.ToNavigationTable(#"Added IsLeaf", {"Name"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf")
    in
        asNav;

/**
 * Request
 */

fetchRequests = (accountId, header) => 
    let
        requests = Json.Document(
            Web.Contents(base_path & prefix & accountId  & "/requests", 
                [
                    Query = [ sort="desc", limit="10", offset="0" ],
                    Headers = header
                ]
            )
        )
    in
        requests;

AutodeskDataConnector.getLastRequest = (requests) =>
    try
        let
            results = requests[results],
            #"Converted to Table" = Table.FromList(results, Splitter.SplitByNothing(), null, null, ExtraValues.Error),
            #"Expanded Column1" = Table.ExpandRecordColumn(
                #"Converted to Table", "Column1", { "id", "lastQueuedAt" }, { "id", "lastQueuedAt" }
            ),
            getFirstAndQueued = Table.SelectRows(
                #"Expanded Column1",
                each [lastQueuedAt] <> null
            ),
            request = getFirstAndQueued{0}
        in
            request
    otherwise
        error Error.Record("Not found", "No data requests for this account in the last 30 days. Users with executive overview access can request data from Autodesk Data Connector-Preview.");

/**
 * Jobs
 */

fetchJobs = (accountId, requestId, header) =>
    let
        jobs = Json.Document(
            Web.Contents(
                base_path & prefix & accountId & "/requests/" & requestId & "/jobs",
                [
                    Headers = header
                ]
            )
        )
    in
        jobs;

AutodeskDataConnector.getLastJob = (jobs) =>
    try
        let

            jobResults = jobs[results],
            jobTable = Table.FromList(
                jobResults, Splitter.SplitByNothing(), null, null, ExtraValues.Error
            ),
            #"Expanded Job Column" = Table.ExpandRecordColumn(
                jobTable, "Column1", { "id" }, { "id" }
            ),
            jobId = #"Expanded Job Column"{0}
        in
            jobId
    otherwise
        error Error.Record("Not found", "No data requests for this account in the last 30 days. Users with executive overview access can request data from Autodesk Data Connector-Preview.");

/**
 * Data listing
 */

fetchDataListing = (accountId, jobId, header) =>
    try
        let
            dataListing = Json.Document(
                Web.Contents(
                    base_path & prefix & accountId & "/jobs/" & jobId & "/data-listing",
                    [
                        Headers = header
                    ]
                )
            )
        in
            dataListing
    otherwise
        error Error.Record("Error", "Unfortunately, your Data Connector job failed. Please re-submit a new Data Connector request to continue.");

AutodeskDataConnector.prepareDataListing = (dataListing) =>
    let
        dataListToTable = Table.FromList(dataListing, Splitter.SplitByNothing(), null, null, ExtraValues.Error),
        #"Expanded Datalist Column" = Table.ExpandRecordColumn(
            dataListToTable, 
            "Column1", 
            {"name"},
            {"Name"}
        ),
        #"Filtered Rows" = Table.SelectRows(
            #"Expanded Datalist Column", 
            each Text.Contains([Name], ".csv")
        ),
        dataListTable = Table.ReplaceValue(
            #"Filtered Rows", ".csv", "", Replacer.ReplaceText, {"Name"}
        )
    in
        dataListTable;

/**
 * CSV
 */

fetchCsvFile = (accountId, jobId, filename, header) =>
    let
        csvUrl = Json.Document(
            Web.Contents(
                base_path & prefix & accountId & "/jobs/" & jobId & "/data/" & filename & ".csv",
                [
                    Headers = header
                ]
            )
        )[signedUrl],
        csv = Csv.Document(Web.Contents(csvUrl, [ ManualCredentials = true ]))
    in
        csv;

getCsv = (accountId, jobId, name, header) =>
    let
        csv = fetchCsvFile(accountId, jobId, name, header),
        withHeaders = Table.PromoteHeaders(csv)
    in
        withHeaders;

/**
 * Get Datalisting
 */
 
getViews = (accountId as text, header) =>
    try
        let
            requests        = fetchRequests(accountId, header),
            requestId       = AutodeskDataConnector.getLastRequest(requests)[id],
            jobs            = fetchJobs(accountId, requestId, header),
            jobId           = AutodeskDataConnector.getLastJob(jobs)[id],
            dataListing     = fetchDataListing(accountId, jobId, header),
            values          = [
                data = AutodeskDataConnector.prepareDataListing(dataListing),
                account = accountId,
                job = jobId
            ]
        in
            values
    otherwise
        error Error.Record("Not found", "No data requests for this account in the last 30 days. Users with executive overview access can request data from Autodesk Data Connector-Preview.");
/**
 * Helper functions for OAuth2: StartLogin, FinishLogin
 */

StartLogin = (resourceUrl, state, display) =>
    let
        authorizeUrl = authorize_uri & "?" & Uri.BuildQueryString([
            response_type = "token",
            client_id = client_id,
            redirect_uri = redirect_uri,
            state = state,
            scope = scope
        ])
    in
        [
            LoginUri = authorizeUrl,
            CallbackUri = redirect_uri,
            WindowHeight = 720,
            WindowWidth = 1024,
            Context = null
        ];

FinishLogin = (context, callbackUri, state) =>
    let
        // parse the full callbackUri, and extract the Query string
        parts = Uri.Parts(Text.Replace(callbackUri, "#", "?"))[Query],
        // if the query string contains an "error" field, raise an error
        result = if (Record.HasFields(parts, {"error", "error_description"})) then
                    error Error.Record(parts[error], parts[error_description], parts)
                 else
                    parts
    in
        result;

/**
 * Common functions
 */

Table.ToNavigationTable = (
    table as table,
    keyColumns as list,
    nameColumn as text,
    dataColumn as text,
    itemKindColumn as text,
    itemNameColumn as text,
    isLeafColumn as text
) as table =>
    let
        tableType = Value.Type(table),
        newTableType = Type.AddTableKey(tableType, keyColumns, true) meta
        [
            NavigationTable.NameColumn = nameColumn,
            NavigationTable.DataColumn = dataColumn,
            NavigationTable.ItemKindColumn = itemKindColumn,
            Preview.DelayColumn = itemNameColumn,
            NavigationTable.IsLeafColumn = isLeafColumn
        ],
        navigationTable = Value.ReplaceType(table, newTableType)
    in
        navigationTable;

/**
 * Data Source UI publishing description
 */
AutodeskDataConnector.Publish = [
    Beta = false,
    Category = "Other",
    ButtonText = { "Autodesk Data Connector - Preview", "Autodesk Data Connector - Preview" },
    LearnMoreUrl = "https://help.autodesk.com/view/BIM360D/ENU/?guid=BIM360D_Insight_data_extractor_html",
    SourceImage = AutodeskDataConnector.Icons,
    SourceTypeImage = AutodeskDataConnector.Icons
];

AutodeskDataConnector.Icons = [
    Icon16 = { Extension.Contents("AutodeskDataConnector16.png"), Extension.Contents("AutodeskDataConnector20.png"), Extension.Contents("AutodeskDataConnector24.png"), Extension.Contents("AutodeskDataConnector32.png") },
    Icon32 = { Extension.Contents("AutodeskDataConnector32.png"), Extension.Contents("AutodeskDataConnector40.png"), Extension.Contents("AutodeskDataConnector48.png"), Extension.Contents("AutodeskDataConnector64.png") }
];
